/*
Module : FluidSimulator.h
Purpose: Provides a C++ implementation of the Fluid Dynamics solver by Jos Stam as described at 
         http://www.dgp.toronto.edu/people/stam/reality/Research/pdf/GDC03.pdf
History: PJN / 24-03-2019 1. Initial public release

Copyright (c) 2019 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise)
when your product is released in binary form. You are allowed to modify the source code in any way you want
except you cannot modify the copyright details at the top of each module. If you want to distribute source
code with your application, then you are only allowed to distribute versions released by the author. This is
to maintain a single distribution point for the source code.

*/


////////////////////////// Macros / Defines ///////////////////////////////////

#pragma once


//////////////////////// Includes /////////////////////////////////////////////

#include <vector>
#include <algorithm>


//////////////////////// Classes //////////////////////////////////////////////

class CFluidSimulator
{
public:
//Constructors / Destructors
  CFluidSimulator(int N, float diff, float dt, float visc) : m_N(N),
                                                             m_diff(diff),
                                                             m_dt(dt),
                                                             m_visc(visc)
  {
    Allocate();
  }
  CFluidSimulator(const CFluidSimulator&) = delete;
  CFluidSimulator(CFluidSimulator&&) = delete;
  ~CFluidSimulator() = default;

//Methods
  CFluidSimulator& operator=(const CFluidSimulator&) = delete;
  CFluidSimulator& operator=(CFluidSimulator&&) = delete;

  static constexpr int IX(int N, int i, int j)
  {
    return i + ((N + 2) * j);
  }

  void vel_step() noexcept
  {
    vel_step(m_u.data(), m_v.data(), m_u_prev.data(), m_v_prev.data());
  }

  void dens_step() noexcept
  {
    dens_step(m_dens.data(), m_dens_prev.data(), m_u.data(), m_v.data());
  }

  void clear()
  {
    const int size = (m_N + 2) * (m_N + 2);
    for (int i=0; i<size; i++)
#pragma warning(suppress: 26446)
      m_u[i] = m_v[i] = m_u_prev[i] = m_v_prev[i] = m_dens[i] = m_dens_prev[i] = 0.0f;
  }

  void SetN(int N)
  {
    m_N = N;
    Allocate();
  }

  int N() const noexcept
  {
    return m_N;
  };

  void SetDiffusion(float diff) noexcept
  {
    m_diff = diff;
  }

  void SetViscosity(float visc) noexcept
  {
    m_visc = visc;
  }

  void SetDt(float dt) noexcept
  {
    m_dt = dt;
  }

//Member variables
  std::vector<float> m_u;
  std::vector<float> m_v;
  std::vector<float> m_u_prev;
  std::vector<float> m_v_prev;
  std::vector<float> m_dens;
  std::vector<float> m_dens_prev;

protected:
//Member variables
  int m_N;
  float m_diff;
  float m_dt;
  float m_visc;

//Methods
  void dens_step(float* x, float* x0, const float* u, const float* v) noexcept
  {
    add_source(x, x0);
    std::swap(x0, x);
    diffuse(0, x, x0, m_diff);
    std::swap(x0, x);
    advect(0, x, x0, u, v);
  }

  void vel_step(float* u, float* v, float* u0, float* v0) noexcept
  {
    add_source(u, u0);
    add_source(v, v0);
    std::swap(u0, u);
    diffuse(1, u, u0, m_visc);
    std::swap(v0, v);
    diffuse(2, v, v0, m_visc);
    project(u, v, u0, v0);
    std::swap(u0, u);
    std::swap(v0, v);
    advect(1, u, u0, u0, v0);
    advect(2, v, v0, u0, v0);
    project(u, v, u0, v0);
  }

#pragma warning(suppress: 26429)
  void add_source(float* x, const float* s) noexcept
  {
    const int size = (m_N + 2) * (m_N + 2);
    for (int i = 0; i < size; i++)
#pragma warning(suppress: 26481)
      x[i] += (m_dt * s[i]);
  }

#pragma warning(suppress: 26429)
  void lin_solve(int b, float* x, const float* x0, float a, float c) noexcept
  {
    for (int k=0; k<20; k++)
    {
      for (int i=1; i<=m_N; i++)
      {
        for (int j=1; j<=m_N; j++)
#pragma warning(suppress: 26481)
          x[IX(m_N, i, j)] = (x0[IX(m_N, i, j)] + a * (x[IX(m_N, i - 1, j)] + x[IX(m_N, i + 1, j)] + x[IX(m_N, i, j - 1)] + x[IX(m_N, i, j + 1)])) / c;
      }
      set_bnd(b, x);
    }
  }

  void diffuse(int b, float* x, const float* x0, float diff) noexcept
  {
    const float a = m_dt * diff * m_N * m_N;
    lin_solve(b, x, x0, a, 1 + 4 * a);
  }

#pragma warning(suppress: 26429)
  void advect(int b, float* d, const float* d0, const float* u, const float* v) noexcept
  {
    const float dt0 = m_dt * m_N;
    for (int i=1; i<=m_N; i++)
    {
      for (int j=1; j<=m_N; j++)
      {
#pragma warning(suppress: 26481)
        float x = i - dt0 * u[IX(m_N, i, j)];
#pragma warning(suppress: 26481)
        float y = j - dt0 * v[IX(m_N, i, j)];
        if (x < 0.5f)
          x = 0.5f;
        if (x > (m_N + 0.5f))
          x = m_N + 0.5f;
        const int i0 = static_cast<int>(x);
        const int i1 = i0 + 1;
        if (y < 0.5f)
          y = 0.5f;
        if (y > (m_N + 0.5f))
          y = m_N + 0.5f;
        const int j0 = static_cast<int>(y);
        const int j1 = j0 + 1;
        const float s1 = x - i0;
        const float s0 = 1 - s1;
        const float t1 = y - j0;
        const float t0 = 1 - t1;
#pragma warning(suppress: 26481)
        d[IX(m_N, i, j)] = s0 * (t0 * d0[IX(m_N, i0, j0)] + t1 * d0[IX(m_N, i0, j1)]) + s1 * (t0 * d0[IX(m_N, i1, j0)] + t1 * d0[IX(m_N, i1, j1)]);
      }
    }
    set_bnd(b, d);
  }

  void Allocate()
  {
    const int size = (m_N + 2) * (m_N + 2);
    m_u.resize(size);
    m_v.resize(size);
    m_u_prev.resize(size);
    m_v_prev.resize(size);
    m_dens.resize(size);
    m_dens_prev.resize(size);
  }

#pragma warning(suppress: 26429)
  void set_bnd(int b, float* x) noexcept
  {
    for (int i=1; i<=m_N; i++)
    {
#pragma warning(suppress: 26481)
      x[IX(m_N, 0, i)] = b == 1 ? -x[IX(m_N, 1, i)] : x[IX(m_N, 1, i)];
#pragma warning(suppress: 26481)
      x[IX(m_N, m_N + 1, i)] = b == 1 ? -x[IX(m_N, m_N, i)] : x[IX(m_N, m_N, i)];
#pragma warning(suppress: 26481)
      x[IX(m_N, i, 0)] = b == 2 ? -x[IX(m_N, i, 1)] : x[IX(m_N, i, 1)];
#pragma warning(suppress: 26481)
      x[IX(m_N, i, m_N + 1)] = b == 2 ? -x[IX(m_N, i, m_N)] : x[IX(m_N, i, m_N)];
    }
#pragma warning(suppress: 26481)
    x[IX(m_N, 0, 0)] = 0.5f * (x[IX(m_N, 1, 0)] + x[IX(m_N, 0, 1)]);
#pragma warning(suppress: 26481)
    x[IX(m_N, 0, m_N + 1)] = 0.5f * (x[IX(m_N, 1, m_N + 1)] + x[IX(m_N, 0, m_N)]);
#pragma warning(suppress: 26481)
    x[IX(m_N, m_N + 1, 0)] = 0.5f * (x[IX(m_N, m_N, 0)] + x[IX(m_N, m_N + 1, 1)]);
#pragma warning(suppress: 26481)
    x[IX(m_N, m_N + 1, m_N + 1)] = 0.5f * (x[IX(m_N, m_N, m_N + 1)] + x[IX(m_N, m_N + 1, m_N)]);
  }

#pragma warning(suppress: 26429)
  void project(float* u, float* v, float* p, float* div) noexcept
  {
    for (int i=1; i<=m_N; i++)
    {
      for (int j=1; j<=m_N; j++)
      {
#pragma warning(suppress: 26481)
        div[IX(m_N, i, j)] = -0.5f * (u[IX(m_N, i + 1, j)] - u[IX(m_N, i - 1, j)] + v[IX(m_N, i, j + 1)] - v[IX(m_N, i, j - 1)]) / m_N;
#pragma warning(suppress: 26481)
        p[IX(m_N, i, j)] = 0;
      }
    }
    set_bnd(0, div);
    set_bnd(0, p);
    lin_solve(0, p, div, 1, 4);
    for (int i=1; i<=m_N; i++)
    {
      for (int j=1; j<=m_N; j++)
      {
#pragma warning(suppress: 26481)
        u[IX(m_N, i, j)] -= 0.5f * m_N * (p[IX(m_N, i + 1, j)] - p[IX(m_N, i - 1, j)]);
#pragma warning(suppress: 26481)
        v[IX(m_N, i, j)] -= 0.5f * m_N * (p[IX(m_N, i, j + 1)] - p[IX(m_N, i, j - 1)]);
      }
    }
    set_bnd(1, u);
    set_bnd(2, v);
  }
};
